#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @author Paul Dechamps
# @date 2022
# Test the blast implementation. The test case is a compressible attached transitional flow.
# Tested functionalities;
# - Time integration.
# - Two time-marching techniques (agressive and safe).
# - Transition routines.
# - Forced transition.
# - Compressible flow routines.
# - Laminar and turbulent flow.
#
# Untested functionalities.
# - Separation routines.
# - Multiple failure case at one iteration.
# - Response to unconverged inviscid solver.

# Imports.

import blast.utils as vutils
import numpy as np

from fwk.wutils import parseargs
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

import math

def cfgInviscid(nthrds, verb):
    import os.path
    # Parameters
    return {
    # Options
    'scenario' : 'aerodynamic',
    'task' : 'analysis',
    'Threads' : nthrds, # number of threads
    'Verb' : verb, # verbosity
    # Model (geometry or mesh)
    'File' : os.path.dirname(os.path.abspath(__file__)) + '/../../models/dart/n0012.geo', # Input file containing the model
    'Pars' : {'xLgt' : 50, 'yLgt' : 50, 'growthRatio': 1.2, 'msTe' : 0.01, 'msLe' : 0.001}, # parameters for input file model
    'Dim' : 2, # problem dimension
    'Format' : 'gmsh', # save format (vtk or gmsh)
    # Markers
    'Fluid' : 'field', # name of physical group containing the fluid
    'Farfield' : ['upstream', 'farfield', 'downstream'], # LIST of names of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    'Wing' : 'airfoil', # LIST of names of physical groups containing the lifting surface boundary
    'Wake' : 'wake', # LIST of names of physical group containing the wake
    'WakeTip' : 'wakeTip', # LIST of names of physical group containing the edge of the wake
    'Te' : 'te', # LIST of names of physical group containing the trailing edge
    'Upstream' : 'upstream',
    # Freestream
    'M_inf' : 0.2, # freestream Mach number
    'AoA' : 2., # freestream angle of attack
    # Geometry
    'S_ref' : 1., # reference surface length
    'c_ref' : 1., # reference chord length
    'x_ref' : .25, # reference point for moment computation (x)
    'y_ref' : 0.0, # reference point for moment computation (y)
    'z_ref' : 0.0, # reference point for moment computation (z)
    # Numerical
    'LSolver' : 'GMRES', # inner solver (Pardiso, MUMPS or GMRES)
    'G_fill' : 2, # fill-in factor for GMRES preconditioner
    'G_tol' : 1e-5, # tolerance for GMRES
    'G_restart' : 50, # restart for GMRES
    'Rel_tol' : 1e-6, # relative tolerance on solver residual
    'Abs_tol' : 1e-8, # absolute tolerance on solver residual
    'Max_it' : 20 # solver maximum number of iterations
    }

def cfgBlast(verb):
    return {
        'Re' : 1e7,                 # Freestream Reynolds number
        'Minf' : 0.2,               # Freestream Mach number (used for the computation of the time step only)
        'CFL0' : 1,                 # Inital CFL number of the calculation
        'Verb': verb,               # Verbosity level of the solver
        'couplIter': 50,            # Maximum number of iterations
        'couplTol' : 1e-4,          # Tolerance of the VII methodology
        'iterPrint': 5,             # int, number of iterations between outputs
        'resetInv' : True,          # bool, flag to reset the inviscid calculation at every iteration.
        'sections' : [0],           # List of sections for boundary layer calculation
        'xtrF' : [None, 0.4],       # Forced transition locations
        'interpolator' : 'Matching' # Interpolator for the coupling
    }

def main():
    # Timer.
    tms = fwk.Timers()
    tms['total'].start()

    args = parseargs()
    icfg = cfgInviscid(args.k, args.verb)
    vcfg = cfgBlast(args.verb)

    tms['pre'].start()
    coupler, isol, vsol = vutils.initBlast(icfg, vcfg)
    tms['pre'].stop()

    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    aeroCoeffs = coupler.run()
    tms['solver'].stop()

    # Display results.
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('      Re        M    alpha       Cl       Cd      Cdp      Cdf       Cm')
    print('{0:6.1f}e6 {1:8.2f} {2:8.1f} {3:8.4f} {4:8.4f} {5:8.4f} {6:8.4f} {7:8.4f}'.format(vcfg['Re']/1e6, isol.getMinf(), isol.getAoA()*180/math.pi, isol.getCl(), vsol.Cdt, vsol.Cdp, vsol.Cdf, isol.getCm()))
    tms['total'].stop()

    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)
    print('SOLVERS statistics')
    print(coupler.tms)

    bl_solution = vutils.getSolution(isol.sec, write=False)[0]
    
    # Test solution
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('Cl', isol.getCl(), 0.230, 5e-2)) # XFOIL 0.2325
    tests.add(CTest('Cd wake', vsol.Cdt, 0.0057, 1e-3, forceabs=True)) # XFOIL 0.00531
    tests.add(CTest('Cd integral', vsol.Cdf + isol.getCd(), 0.0058, 1e-3, forceabs=True)) # XFOIL 0.00531
    tests.add(CTest('Cdf', vsol.Cdf, 0.0047, 1e-3, forceabs=True)) # XFOIL 0.00084, Cdf = 0.00447
    tests.add(CTest('xtr_top', vsol.getAverageTransition(0), 0.196, 0.03, forceabs=True)) # XFOIL 0.1877
    tests.add(CTest('xtr_bot', vsol.getAverageTransition(1), 0.40, 0.01, forceabs=True)) # XFOIL 0.4998
    tests.add(CTest('error dStar', np.linalg.norm(bl_solution['deltaStar'] - bl_solution['theta']*bl_solution['H']), 0., 0., forceabs=True))
    tests.add(CTest('Iterations', len(aeroCoeffs['Cl']), 20, 0, forceabs=True))
    tests.run()

    # Plot results
    if not args.nogui:
        cpv = np.loadtxt('Cp_viscous.dat', skiprows=1, delimiter=',')
        cpi = np.loadtxt('Cp_inviscid.dat', skiprows=1, delimiter=',')
        cpv_xf = np.loadtxt('/'.join(__file__.split('/')[:-3])+'/models/references/n0012_2deg_xf_cpVisc.dat', skiprows=1)
        cpi_xf = np.loadtxt('/'.join(__file__.split('/')[:-3])+'/models/references/n0012_2deg_xf_cpInv.dat', skiprows=1)
        bl_xf = np.loadtxt('/'.join(__file__.split('/')[:-3])+'/models/references/n0012_2deg_xf_bl.dat', skiprows=1)

        from matplotlib import pyplot as plt
        plt.figure()
        plt.plot(cpv[:, 0], cpv[:, 3], color='firebrick', lw=2, label='BLASTER - viscous')
        plt.plot(cpv_xf[:, 0], cpv_xf[:, 1], color='royalblue', lw=2, label='XFOIL - viscous')
        plt.plot(cpi[:, 0], cpi[:, 3], color='firebrick', lw=1, label='BLASTER - inviscid', linestyle='--')
        plt.plot(cpi_xf[:, 0], cpi_xf[:, 1], color='royalblue', lw=1, label='XFOIL - inviscid', linestyle='--')
        plt.gca().invert_yaxis()
        plt.legend(frameon=False)
        plt.xlabel('$x/c$')
        plt.ylabel('$c_p$')
        plt.draw()

        plt.figure()
        plt.plot(bl_solution['x'], bl_solution['cf']*bl_solution['ue']*bl_solution['ue'], color='firebrick', lw=2, label='BLASTER')
        plt.plot(bl_xf[:, 1], bl_xf[:, 6], color='royalblue', lw=2, label='XFOIL')
        plt.legend(frameon=False)
        plt.xlim([0, 1])
        plt.xlabel('$x/c$')
        plt.ylabel('$c_f$')
        plt.draw()

        plt.show()

    # eof
    print('')

if __name__ == "__main__":
    main()
