#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @author Paul Dechamps
# @date 2024
# Test the blaster adjoint implementation.

# Imports.
import numpy as np

from fwk.wutils import parseargs
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

import fwk
from fwk.testing import *
from fwk.coloring import ccolors
import numpy as np
import os

from matplotlib import pyplot as plt
from blast.api.blaster_api import init_blaster
import blast.utils as vutils

def cfgInviscid(nthrds, verb):
    import os.path
    # Parameters
    return {
    # Options
    'scenario' : 'aerodynamic',
    'task' : 'optimization',
    'Threads' : nthrds, # number of threads
    'Verb' : verb, # verbosity
    # Model (geometry or mesh)
    'File' : os.path.dirname(os.path.abspath(__file__)) + '/../models/dart/n0012.geo', # Input file containing the model
    'Pars' : {'xLgt' : 50, 'yLgt' : 50, 'msF' : 10, 'msTe' : 0.01, 'msLe' : 0.01}, # parameters for input file model
    'Dim' : 2, # problem dimension
    'Format' : 'gmsh', # save format (vtk or gmsh)
    # Markers
    'Fluid' : 'field', # name of physical group containing the fluid
    'Farfield' : ['upstream', 'farfield', 'downstream'], # LIST of names of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    'Wing' : 'airfoil', # LIST of names of physical groups containing the lifting surface boundary
    'Wake' : 'wake', # LIST of names of physical group containing the wake
    'WakeTip' : 'wakeTip', # LIST of names of physical group containing the edge of the wake
    'Te' : 'te', # LIST of names of physical group containing the trailing edge
    'Upstream' : 'upstream',
    # Freestream
    'M_inf' : 0.2, # freestream Mach number
    'AoA' : 2., # freestream angle of attack
    # Geometry
    'S_ref' : 1., # reference surface length
    'c_ref' : 1., # reference chord length
    'x_ref' : .25, # reference point for moment computation (x)
    'y_ref' : 0.0, # reference point for moment computation (y)
    'z_ref' : 0.0, # reference point for moment computation (z)
    # Numerical
    'LSolver' : 'SparseLU', # inner solver (Pardiso, MUMPS or GMRES)
    'G_fill' : 2, # fill-in factor for GMRES preconditioner
    'G_tol' : 1e-5, # tolerance for GMRES
    'G_restart' : 50, # restart for GMRES
    'Rel_tol' : 1e-6, # relative tolerance on solver residual
    'Abs_tol' : 1e-8, # absolute tolerance on solver residual
    'Max_it' : 20, # solver maximum number of iterations
    }

def cfgBlast(verb):
    return {
        'Re' : 1e7,                     # Freestream Reynolds number
        'Minf' : 0.2,                   # Freestream Mach number (used for the computation of the time step only)
        'CFL0' : 1,                     # Inital CFL number of the calculation
        'Verb': verb,                   # Verbosity level of the solver
        'couplIter': 100,               # Maximum number of iterations
        'couplTol' : 1e-2,              # Tolerance of the VII methodology
        'iterPrint': 20,                # int, number of iterations between outputs
        'resetInv' : True,              # bool, flag to reset the inviscid calculation at every iteration.
        'sections' : [0],               # List of sections for boundary layer calculation
        'xtrF' : [-1, -1],              # Forced transition locations
        'interpolator' : 'Matching',    # Interpolator for the coupling
    }

def main():

    tms = fwk.Timers()

    tms['total'].start()

    # Parse agrs
    args = parseargs()
    verb = args.verb
    nthrds = args.k

    icfg = cfgInviscid(nthrds, verb)
    cfg = cfgBlast(verb)

    obj = init_blaster(cfg, icfg, iSolverName='DART', task='optimization')
    coupler, isol, vsol, adj = obj['coupler'], obj['isol'], obj['vsol'], obj['adj']

    # Forward problem
    coupler.run()

    # Adjoint problem
    isol.adjointSolver.run()
    adj.run()
    tms['total'].stop()

    # Make the api crash
    params = {
        'Re': -1,
        'Minf': -1,
        'CFL0': -1,
        'couplIter': -1,
        'couplTol': -1,
        'iterPrint': -1
    }

    # Loop over the parameters
    for param, value in params.items():
        original = cfg[param]
        try:
            cfg[param] = value
            obj = init_blaster(cfg, icfg, iSolverName='DART', task='optimization')
            raise AssertionError(f'API initialized with {param} = ', cfg[param])
        except AssertionError as e:
            raise RuntimeError(e)
        except:
            print(ccolors.ANSI_GREEN + f'API crashed with {param} = ' + str(cfg[param]) + ccolors.ANSI_RESET)
            cfg[param] = original

    print('Statistics')
    print(tms)
    print(ccolors.ANSI_BLUE + 'PyTesting' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('Cl', isol.getCl(), 0.234, 1e-2, forceabs=True))
    tests.add(CTest('Cd', isol.getCd()+vsol.Cdf, 0.005, 1e-2, forceabs=True))
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()